#! /usr/bin/env python
# -*- coding: utf8 -*-

#
# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
#


__doc__ = """

This module is responsible for parsing xml nufw firewall desccription, and mapping it into objects.

"""
import xml.dom.minidom
from IPy import IP
from nupyf.net_defs import INTERNET
from nupyf.nuacl import Rule

ZONE_EXT = "external"
ZONE_INT = "internal"
ZONE_VPN = "vpn"

DFT_GATEWAY = "dftgateway"

class Net:
    """Maps <net> tag from a network description xml file (nuface interface)
    """
    def __init__(self, name, addr, zone='', type='', mark='', remote='', id=None, enabled=1):
        if(zone==''):
            self.__zone = ZONE_INT
        else:
            self.__zone = zone
        self.__conns = []
        if not addr:
            addr = '0.0.0.0/0'
        self.__addr = IP(addr)
        self.__name = name
        self.__type = type
        self.__mark = mark
        self.__remote = remote
        self.local_id = None
        self.__enabled = enabled
        self.__id = 0
        if id is not None: self.__id = int(id)

    def add_conn(self, dir, fwid, iface, addr = None, gw = None,
            dftgw = None, key=None, side=None):
        h = {'direct':dir, 'fwid':fwid, 'iface':iface, 'addr': addr,
                'key': key, 'side': side}
        #if(len(args)==1):
        #    h['gateway'] = args[0]
        if gw:
            h['gateway'] = gw
        if dftgw:
            h[DFT_GATEWAY] = dftgw
        self.__conns.append(h)

    def get_conn(self, fwid):
        for conn in self.__conns:
            if conn['fwid']==fwid:
                return conn

    def iter_conns(self):
        for index in range(0, len(self.__conns)):
            yield self.__conns[index]

    def get_conns(self):
        """get connexions"""
        return self.__conns

    #getters
    def netmask(self):
        """get the netmask of the network
        """
        return self.__addr.netmask()

    def ip(self):
        """return the string representing the addr of the Net object"""
        return str(self.__addr)

    def addr(self):
        return self.__addr

    def name(self):
        return self.__name

    def zone(self):
        return self.__zone

    def type(self):
        return self.__type

    def mark(self):
        return self.__mark

    def remote(self):
        return self.__remote

    # used by nuconf, vpn.php
    def is_enabled(self):
        return self.__enabled

    def is_vpn(self):
        """Checks if the network is a vpn """
        return self.__mark != '' or self.local_id != ''

    def id(self):
        """Get network id"""
        return self.__id

    def __str__(self):
        return 'name: %s, addr:%s, id:%s, zone:%s' % (
            self.__name, self.__addr, self.__id, self.__zone)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if self.__id == other.__id and self.__addr == other.__addr:
            return True
        return False

    def __ne_(self, other):
        return not self.__eq__(other)

class Interface:
    """An interface has an ip addresse and a list of networks attached to it.
    """
    def __init__(self, id, name):
        self.__id = id
        # Reseaux lies à une interfaces
        self.__directnets = []
        self.__addr = {}
        self.__name = name
        # Réseaux routés
        self.__distnets = []
        self.__nets = []
        self.__gateways = {}
        self.__addr2net = {}
        self.__distnet2addr = {}

    def name(self) :
        return self.__name

    def add_addr(self, id, addr):
        assert id not in self.__addr
        self.__addr[id] = IP(addr)

    def add_directnet(self, net, vpn=False):
        addr_in_net = False
        for addr in self.__addr.values():
            if addr in net.addr():
                addr_in_net = True
                self.__addr2net[addr] = net
        if not addr_in_net and not vpn:  # FIXME: Fix previous loop.
            raise ValueError("Network %s (%s) doesn't match to any firewall" % (net.name(), net.addr()))

        self.__directnets.append(net)

    def add_distnet(self, net, gateway):
        self.__distnets.append(net)
        gateway_net = None
        related_addr = None
        for addr, direct in self.__addr2net.iteritems():
            direct = direct.addr()
            if gateway in direct and (gateway_net is None or direct < gateway_net):
                gateway_net = direct
                related_addr = addr
        if related_addr  is None:
            raise ValueError("Unable to find the route for the address %s" % net.addr())
        self.__distnet2addr[net.addr()] = related_addr

    def add_net(self, net):
        self.__nets.append(net)

    def distnets(self):
        return iter(self.__distnets)

    def allnets(self):
        """return all the network of the interface
        """
        return iter(self.__nets)

    def addr(self):
        return map(str, self.__addr.values())

    def addr_values(self):
        return self.__addr.values()

    def get_addr_by_id(self, id):
        return str(self.__addr[id])

    def id(self):
        return self.__id

    def directnets(self):
        return self.__directnets[:]

    def self_net(self):
        """return a Net object that represents the ip addr of the interface"""
        return Net(self.__name, None)

    def get_net_by_addr(self, addr):
        net =  self.__addr2net.get(addr)
        if net is None:
            raise KeyError('%r not found in %s' %(addr, self.__addr2net))
        return net

    def get_local_ip_connected_to_distnet(self, distnet):
        assert(distnet in self.__distnets)
        return self.__distnet2addr[distnet.addr()]

    def __repr__(self):
        return "<Interface name=%r id=%s addr=%s>" % (
            self.__name, self.__id, self.__addr)

#End class Interface
class RuleList(object):
    """List of rules
    """
    def __init__(self):
        self.rules = {}
        self.rule_list = []
        self.bad = 0

    def is_bad(self):
        return self.bad == 1

    def add_rule(self, rule, order):
        self.rule_list.append((order, rule))

    def ordered_rules(self):
        self.rule_list.sort()
        return self.rule_list[:]


class FWNetConf(object):
    """DNS configuration for firewall
    """
    def __init__(self):
        self.__dns_search = []
        self.__dns_nameservers = []
        pass

    def add_dns_search(self, resolv):
        """add a search item for dns configuration.
        parameters: resolv = str
        """
        self.__dns_search.append(resolv)

    def add_dns_nameserver(self, server):
        """Add a dns namserver.
        parameters: server = str. the ip addr of the namserver
        """
        self.__dns_nameservers.append(server)

    def get_dns_search(self):
        """Return list of dns resolv domains
        """
        return self.__dns_search
    def get_dns_nameservers(self):
        """Return the list of dns nameservers
        """
        return self.__dns_nameservers

    def get_dns_conf(self):
        """Get dns configuration. Returns (resolv list, nameservers list)
        """
        return (self.__dns_search, self.__dns_nameservers)
#End class FWNetConf

class FW:
    """FW object is a firewall with:
- Interfaces
- Nets
- Rules
Rules are stored in chains corresponding to their source/destination ip addresses.
    """
    def __init__(self, id):
        self.__ifaces = []
        self.__nets = []
        self.__id = id
        self.__hrules = {}
        self.__allrules = []
        self.__nat_rules = {'snat': [], 'dnat': [], 'pnat': []}
        self.__chains = []
        self.__ext_net = None
        self.__vpns = {}
        self.__netconf = None
        self.queue = None
        self.auth_ext = False #authenticate packets from internet?
        self.same_iface = False #don't generate rules with same input and output interface?
        self.no_check_net = False # don't check inclusion and overlaping on enterprise networks


    def id(self):
        """get firewall id"""
        return self.__id

    def has_dft_net(self):
        return self.__ext_net is not None

    def add_iface(self, iface):
        self.__ifaces.append(iface)

    def ifaces(self):
        return iter(self.__ifaces)

    def nb_ifaces(self):
        return len(self.__ifaces)

    def add_net(self, net, conn = None):
        self.__nets.append(net)
        #More logical than net.zone():
        if conn is not None and DFT_GATEWAY in conn:
            self.__ext_net = net
        if net.is_vpn():
            self.__vpns[net.id()] = net


    #  Checks if a network is the default route network for the firewall
    def is_default_net(self, net):
        """Checks if a network is the default route network for the firewall
           net: Net object, the network to be checked
        """
        if not self.__ext_net:
            return False
        return net.id() == self.__ext_net.id()

    def is_normal_net(self, net):
        """Check if a network is not default route nor vpn"""
        return (net.is_vpn() is not True and self.is_default_net(net) is not True)

    def routes(self):
        s = ''
        for iface in self.__ifaces:
            for n in iface.distnets():
                h = n.get_conn(self.__id)
                s+="%s\t%s\n" %(str(n.addr().net()), h['router'])
        return s

    def __add_chain_for_rules_if_not_exist(self, chain):
        if chain not in self.__hrules:
            self.__hrules[chain] = RuleList()

    def find_zone(self, ip, mark):
        """Find the zone of the firewall which contains the string ip address "ip" and/or
        the string mark
        """
        if mark != '' and mark in self.__vpns:
            return ZONE_VPN
        for net in self.__nets:
            if net.zone != ZONE_EXT and ip in net.addr():
                return net.zone()
        return ZONE_EXT

    def is_correct_network(self, network):
        """Check if network is correct
        network should be included in one enterprise network
        if not: network must not overlap with enterprise network
        """
        net = IP(network)
        l = self.cached_entreprise_nets()
        included = 0
        for net2 in l:
            if net in net2:
                included = 1
        if included:
            return True
        else:
            for nete in l:
                if nete.overlaps(net):
                    return False
            return True

    def get_local_ip_connected_to(self, remote_ip):
        """Returns ip address of fw which can route to remote_ip
        """
        remote = IP(remote_ip)
        internet = IP('0.0.0.0/0')
        to_internet = (remote == internet)
        best_candidate = None
        candidate_dict = None
        for iface in self.__ifaces:
            for addr in iface.addr_values():
                net = iface.get_net_by_addr(addr)
                if not to_internet and remote in net.addr() and (best_candidate is None or net.addr() < best_candidate):
                    best_candidate = net.addr()
                    candidate_dict = {'iface':iface, 'addr':addr}
                    continue
                if to_internet and DFT_GATEWAY in net.get_conn(self.__id):
                    return {'iface':iface, 'addr':addr}
            for net in iface.distnets():
                addr = iface.get_local_ip_connected_to_distnet(net)
                conn = net.get_conn(self.__id)
                if not to_internet and remote in net.addr() and (best_candidate is None or net.addr() < best_candidate):
                    best_candidate = net.addr()
                    candidate_dict = {'iface':iface, 'addr':addr}
                    continue
                assert(not (to_internet and DFT_GATEWAY in conn))
        if candidate_dict is None:
            raise ValueError("No route to %s" % remote_ip)
        return candidate_dict

    def chain(self, rule, ip_src, ip_dst, zone_src = '', zone_dst = '', mark_src = '', mark_dst = ''):
        """
        rule: Rule or GenNat object.

        Returns the fw chain that fits ip_src/ip_dst/zone_src/zone_dst/mark
        """
        # FIXME: Do not use anymore zone attribute for elements. (OK for networks)
        if isinstance(rule, Rule):
            acl_type = rule.acl.acl_type
        else:
            acl_type = "NAT"
        chp1 = None
        chp2 = None
        if1 = None
        if2 = None
        ip_src = IP(ip_src)
        ip_dst = IP(ip_dst)
        #Let's find the right network. no need of rules zones.
        for iface in self.__ifaces:
            net_id_by_broadcast = dict([ (net.addr().broadcast(), net.id()) for net in iface.directnets() ])
            iface_addr = iface.addr_values()
            for net in iface.allnets():
                # got source network?
                if ip_src in net.addr():
                    # we want the smallest network
                    if chp1 is not None and net.addr() <= chp1.addr():
                        chp1 = net
                    elif chp1 is None:
                        chp1 = net
                # got destination network?
                if ip_dst in net.addr():
                    # we want the smallest network
                    if chp2 is not None and net.addr() <= chp2.addr():
                        chp2 = net
                    elif chp2 is None:
                        chp2 = net
            if acl_type != "FORWARD":
                #INPUT or OUTPUT: all interfaces, default included
                #create fake network for input/output

                # INPUT
                iface_net_id = None
                if ip_src in iface_addr :
                    iface_net_id = iface.get_net_by_addr(ip_src).id()
                else:
                    iface_net_id = net_id_by_broadcast.get(ip_src)
                if iface_net_id is not None:
                    chp1 = Net('IF', ip_src.strNormal(), type = 'ipv4', zone = ZONE_INT, id=(-1))
                    chp1.add_conn('0', self.__id, iface.id())
                    if1 = iface

                # OUTPUT
                iface_net_id = None
                if ip_dst in iface_addr :
                    iface_net_id = iface.get_net_by_addr(ip_dst).id()
                else:
                    iface_net_id = net_id_by_broadcast.get(ip_dst)
                if iface_net_id is not None:
                    chp2 = Net('IF', ip_dst.strNormal(), type = 'ipv4', zone = ZONE_INT, id=(-1))
                    chp2.add_conn('0', self.__id, iface.id())
                    if2 = iface
        #have we find what we want? if not, got to the evil default route.
        #FIXME NOINT what todo if no ext_net?
        if not chp1:
            chp1 = self.__ext_net
        if not chp2:
            chp2 = self.__ext_net
        if not if1:
            if1 = self.get_iface_by_net(chp1)
        if not if2:
            if2 = self.get_iface_by_net(chp2)
        return {
            'chain': '%s-%s' % (chp1.name(), chp2.name()),
            'netsrc': chp1,
            'netdst': chp2,
            'ifsrc': if1,
            'ifdst': if2}

    # TODO: get INPUT:OUTPUT chains with inout = 1
    def chains(self, all = 0, inout = 0):
        """
        Create a list of firewall's chains.

        all = 1 : all networks are considered , including distant ones
        inout = 1: chains for input/output are also generated
        """
        chains = []
        chains_end = []
        for iface in self.__ifaces:
            for net in iface.directnets():
                if( (all == 0 and self.is_normal_net(net)) or all == 1):
                    chains.append({'name':net.name(), 'net':net, 'zone':None, 'iface':iface})

            for n in iface.distnets():
                if(all == 1):
                    chains.append({'name':n.name(), 'net':n, 'zone':None, 'iface':iface})

            if inout == 1:
                # create chain for interface IP and broadcast IP
                for n in iface.allnets():
                    choutput = {
                        'name': 'IF-'+n.name(),
                        'net_dst': n,
                        'type': "output",
                        'if_src': iface,
                        'if_dst': iface,
                    }
                    self.__add_chain_for_rules_if_not_exist('IF-'+n.name())

                    chinput = {
                        'name': n.name()+'-IF',
                        'net_src': n,
                        'type':"input",
                        'if_src': iface,
                        'if_dst':iface,
                    }
                    self.__add_chain_for_rules_if_not_exist('%s-IF'%(n.name()))
                    chains_end.extend([choutput, chinput])

        tmpres = self._chains_dict(chains, chains)
        tmpres.extend(chains_end)
        return tmpres

    def _chains_dict(self, chains1, chains2):
        res = []
        for chain1 in chains1:
            for chain2 in chains2:
                if not self.same_iface and chain1['iface'] == chain2['iface']:
                    continue
                chain_name = chain1['name']+'-'+chain2['name']
                item = {
                    'name': chain_name,
                    'net_src': chain1['net'],
                    'net_dst': chain2['net'],
                    'type': "forward" ,
                    'if_src': chain1['iface'],
                    'if_dst': chain2['iface'],
                }
                res.append(item)
                if chain_name not in self.__hrules:
                    self.__hrules[chain_name] = RuleList()
        return res

    def entreprise_nets(self):
        l = []
        for iface in self.__ifaces:
            for n in iface.allnets():
                if self.is_normal_net(n):
                    l.append(str(n.addr()))
        return l

    def cached_entreprise_nets(self):
        if hasattr(self, "_FW__enterprise_nets"):
            return self.__enterprise_nets
        l = []
        for iface in self.__ifaces:
            for n in iface.allnets():
                #if(n.zone()==ZONE_INT): #no more zone
                if self.is_normal_net(n):
                    l.append(n.addr())
        self.__enterprise_nets = l
        return l

    def add_rule(self, r):
        """Add a rule to the firewall into the right chain
        """
        if r.src():
            elt_src = r.elt_src()
        else:
            elt_src = None
        if r.dest():
            elt_dst = r.elt_dest()#find and set the zones if they're not set
        else:
            elt_dst = None
        assert (elt_src and elt_src.get('numnet') or (elt_dst and elt_dst.get('numnet')))
        if elt_src and elt_src.get('numnet'):
            netsrc = IP(elt_src.get('numnet'))
        else:
            netsrc = self.get_local_ip_connected_to(elt_dst.get('numnet'))['addr']
        if elt_dst and elt_dst.get('numnet'):
            netdst = IP(elt_dst.get('numnet'))
        else:
            netdst = self.get_local_ip_connected_to(elt_src.get('numnet'))['addr']
        # check inclusion and overlaping for dest/src networks with enterprise networks
        if not self.no_check_net:
            if netsrc != INTERNET and not self.is_correct_network(str(netsrc)):
                raise ValueError("network not correct: %s"%(elt_src.get('numnet')))
            if netdst != INTERNET and not self.is_correct_network(str(netdst)):
                raise ValueError("network not correct: %s"%(elt_dst.get('numnet')))
        if r.src():
            zone_src = r.elt_src().get('zone')
            mark_src = r.elt_src().get('mark')
        else:
            zone_src = mark_src = None
        if r.dest():
            zone_dest = r.elt_dest().get('zone')
            mark_dest = r.elt_dest().get('mark')
        else:
            zone_dest = mark_dest = None
        ch = self.chain(r, str(netsrc), str(netdst),
            zone_src=zone_src,
            zone_dst=zone_dest,
            mark_src=mark_src,
            mark_dst=mark_dest)
        #cpl used to get rule order in chain
        chain_cpl_id = (ch['netsrc'].id(), ch['netdst'].id())
        r_order = None
        if r.sort_order:
            if chain_cpl_id in r.sort_order:
                r_order = r.sort_order[chain_cpl_id]
        else:
            r_order = r.aclid
        r.netsrc = ch['netsrc']
        r.netdst = ch['netdst']
        if(ch['chain'] not in self.__hrules):
            self.__hrules[ch['chain']] = RuleList()
            self.__hrules[ch['chain']].bad = 1
        # add rule only if it has an order in the bichain
        if r_order is not None:
            self.__hrules[ch['chain']].add_rule(r, r_order)
            self.__allrules.append(r)
        else:
            raise ValueError("acl %s: no descsort in %s"%(r.aclid, str(chain_cpl_id)))

    def set_nat_rules(self, snat, dnat, pnat):
        """Set NAT rules for FW object
        """
        # set rules interfaces
        if self.__ext_net:
            ext_iface = self.get_iface_by_net(self.__ext_net).name()
        for r in snat:
            d = self.chain(r, r._to.strNormal(), r._to.strNormal())
            r._iface = d['ifdst'].name()
        # nat/pnat: see if destination address is set
        for r in dnat:
            if r._src:
                d = self.chain(r, r._src.strNormal(), r._src.strNormal())
                r._iface = d['ifdst'].name()
            else:
                r._iface = ext_iface
        for r in pnat:
            if r._dst:
                d = self.chain(r, r._dst.strNormal(), r._dst.strNormal())
                r._iface = d['ifdst'].name()
        self.__nat_rules = {'snat': snat, 'dnat': dnat, 'pnat': pnat}

    def find_network(self, addr):
        """Given a ip address, find the network that this ip belongs to"""
        for iface in self.__ifaces:
            for net in iface.allnets():
                if not self.is_default_net(net):
                    if addr in net.addr():
                        return net
        return self.__ext_net#FIXME NOINT

    def rule_dst_is_src(self, r):
        """Test if a rule comes from and goes to the same network """
        src = r.elt_src(); net_src = None
        dst = r.elt_dest(); net_dst = None
        if net_src is None:
            net_src = self.find_network(src.get('numnet'))
        if net_dst is None:
            net_dst = self.find_network(dst.get('numnet'))
        if self.is_default_net(net_src) and self.is_default_net(net_dst): return True
        if net_dst.addr() in net_src.addr() and net_src.addr() in net_dst.addr(): return True
        return False


    def rule_inout(self, rule):
        """Test if the rule has its src/dest = one of the fw ifaces"""
        return self.rule_input(rule) or self.rule_output(rule)

    def rule_input(self, rule):
        """Test if the rule has its dest = one of the fw ifaces"""
        if rule.elt_dest().get('is_local'):
            # Easy case : elt_dest is a local_ipv4
            return True
        iptest = rule.elt_dest().get('numnet')

        iptest = IP(iptest)
        for iface in self.__ifaces:
            if iptest in [ net.addr().broadcast() for net in iface.directnets() ] :
                return True
            for ifaceaddr in iface.addr():
                iface_addr = IP(ifaceaddr)
                if iface_addr == iptest: return True
        return False

    def rule_output(self, rule):
        """Test if the rule has its source = one of the fw ifaces or network broadcast"""
        if rule.elt_src().get('is_local'):
            # Easy case : elt_src is a local_ipv4
            return True
        iptest = rule.elt_src().get('numnet')
        iptest = IP(iptest)
        for iface in self.__ifaces:
            if iptest in [ net.addr().broadcast() for net in iface.directnets() ] :
                return True
            for ifaceaddr in iface.addr():
                iface_addr = IP(ifaceaddr)
                if iface_addr == iptest: return True
        return False

    def chain_input(self, chain):
        """Test if the chain has its dest = one of the fw iface address or network broadcast"""
        return chain['type'] == "input"

    def chain_output(self, chain):
        """test if the chain has its source = one of the fw iface address or network broadcast"""
        return chain['type'] == "output"

    def chain_inout(self, chain):
        """test if the chain has its src or dst = one of the fw ifaces address or network broadcast"""
        return self.chain_input(chain) or self.chain_output(chain)

    def is_iface_addr_broadcast(self, addr):
        """Test if the string addr is one of the fw ifaces addresses or iface broadcast addr"""
        for iface in self.__ifaces:
            if addr in [ net.addr().broadcast().strNormal() for net in iface.directnets() ]:
                return True
            for ifaceaddr in iface.addr():
                if ifaceaddr == addr: return True
        return False

    def hrules(self):
        return self.__hrules

    def allrules(self):
        return self.__allrules

    def get_nat_rules(self):
        return [self.__nat_rules['snat'], self.__nat_rules['dnat'], self.__nat_rules['pnat']]

    def set_netconf(self, net_conf):
        """set netconf member of FW object
        """
        self.__netconf = net_conf

    def get_netconf(self):
        """Get Net Conf of Firewall
        returns: FwNetConf object
        """
        return self.__netconf

    def get_iface_by_net(self, net):
        """ Get iface corresponding to a specific network"""
        for tmpiface in self.__ifaces:
            if net in tmpiface.directnets():
                return tmpiface
        return None

    def get_iface_by_net_addr(self, net_addr):
        net_addr = IP(net_addr)
        for iface in self.__ifaces:
            if iface.directnets()[0].addr() == net_addr:
                # TODO: Loop over directnets().
                return iface
            for net in iface.distnets():
                if net.addr() == net_addr:
                    return iface
        return None

    def get_iface_by_id(self, id):
        """get firewall iface referenced by its id in xml"""
        for iface in self.__ifaces:
            if int(iface.id()) == int(id):
                return iface
        return None

    def get_net_by_id(self, id):
        """Find a network identified by its ID"""
        id = int(id)  # In case the parameter is a string or a unicode.
        for net in self.__nets:
            if net.id() == id:
                return net
        return None

    def get_ext_net(self):
        """get default network of firewall"""
        return self.__ext_net

    def get_vpns(self):
        """Returns vpn networks of firewall"""
        return self.__vpns

    def get_local_ips(self):
        """Returns ip adresses of firewall"""
        ips = []
        for iface in self.__ifaces:
            ips.extend(map(str, iface.addr()))
        return ips
#End class FW


def _xml_parse_fwnetconf(xmlnode):
    """returns a FWNetConf object
    parameters: xmlnode = XmlNode (fw node)
    """
    ret = FWNetConf()
    dns = xmlnode.getElementsByTagName('dns')
    if not dns: return None
    for mynode in dns[0].childNodes:
        if mynode.nodeType == xml.dom.minidom.Node.ELEMENT_NODE and mynode.nodeName == 'search':
            if mynode.hasAttribute('val'):
                ret.add_dns_search(mynode.getAttribute('val'))
        elif mynode.nodeType == xml.dom.minidom.Node.ELEMENT_NODE and mynode.nodeName == 'nameserver':
            if mynode.hasAttribute('val'):
                ret.add_dns_nameserver(mynode.getAttribute('val'))
    return ret


def parse(file):
    """Parse a file that follows nuface firewall description xml format.
Returns a list of FW objects.
    """
    fwhash = {}
    doc = xml.dom.minidom.parse(file)
    netl = {}

    # Parse networks
    eltnets = doc.getElementsByTagName('nets')
    eltsnet = eltnets[0].childNodes
    for eltnet in eltsnet:
        if(eltnet.nodeType == xml.dom.minidom.Node.ELEMENT_NODE and eltnet.nodeName=="net"):
            if eltnet.hasAttribute('enabled'):
                enabled = (eltnet.getAttribute('enabled') == u'1')
            else:
                enabled = True
            cnet = Net(eltnet.getAttribute('name'),
                eltnet.getAttribute('addr'),
                type=eltnet.getAttribute('type'),
                mark=eltnet.getAttribute('mark'),
                remote=eltnet.getAttribute('remote'),
                id=eltnet.getAttribute('id'),
                enabled=enabled,
                zone=None)
            cnet.local_id = eltnet.getAttribute('local_id')
            if(eltnet.hasChildNodes()):
                for conn in eltnet.getElementsByTagName('connection'):
                    fwid = conn.getAttribute('fwid')
                    direct = conn.getAttribute('direct')
                    iface = conn.getAttribute('iface')
                    addr = conn.getAttribute('addr')
                    key = conn.getAttribute('key')
                    side = conn.getAttribute('side')
                    if(direct=='1'):
                        cnet.add_conn(direct, fwid, iface, dftgw = conn.getAttribute(DFT_GATEWAY))
                    else:
                        cnet.add_conn(direct, fwid, iface, addr, gw =
                                conn.getAttribute("gateway"), dftgw =
                                conn.getAttribute(DFT_GATEWAY), key=key,
                                side=side)
                    # Put new found net into global nets list
                    if(fwid not in netl):
                        netl[fwid] = [cnet]
                    else:
                        netl[fwid].append(cnet)

    # Parse firewalls
    for elt in doc.getElementsByTagName("fw"):
        fwid = elt.getAttribute('id')
        eltsinterface = elt.getElementsByTagName('interface')
        myfw = FW(fwid)
        # set queue number (NFQUEUE)
        if elt.hasAttribute('queue'):
            queue = elt.getAttribute('queue')
            if queue.isdigit():
                myfw.queue = queue
        myfw.set_netconf(_xml_parse_fwnetconf(elt))#get net conf attributes
        for interface in eltsinterface:
            myiface = Interface(interface.getAttribute('id'), interface.getAttribute("name"))
            eltsaddr = interface.getElementsByTagName('address')
            for addr in eltsaddr:
                myiface.add_addr(addr.getAttribute('id'), addr.getAttribute("addr"))
            myfw.add_iface(myiface)
            distnets = []
            for net in netl[fwid]:
                for conn in net.iter_conns():
                    if (conn['fwid']==fwid and conn['iface']==myiface.id()):
                        if conn.get('gateway'):  # Routed network.
                            distnets.append((net, conn['gateway']))
                        else:  # Direct net or VPN.
                            myiface.add_directnet(net,
                                    vpn=conn.get('direct') == '0')
                        myiface.add_net(net)
                        break # TODO: we have the connection
                myfw.add_net(net, conn = conn)
            for dnet, gateway in distnets:
                myiface.add_distnet(dnet, gateway)
        fwhash[fwid] = myfw
    return fwhash


