# Copyright(C) 2005 INL
# Written by Jean Gillaux <jean@inl.fr>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
from nupyf.nuerror import LinkError, LoopError

class Elt:
    def __init__(self):
        self.ID = None
        self.__hash={'id_son': '0'}

    def get(self,key):
        """Get value of element attibute designed by <key>
           empty chain is returned if the attribute doesn't exist
        """
        return self.__hash.get(key,'').encode("utf-8")

    def set(self,key,value):
        """Set the key attribute element with value
        """
        self.__hash[key]=value

    def keys(self):
        return self.__hash.keys()

    def values(self):
        return self.__hash.values()

    def merge_elt(self,other):
        """merge attributes of other element into self element
        """
        for k in other.keys():
            so = other.get(k)
            sv = self.get(k)
            if (sv==''):
                self.set(k,so)

    def __repr__(self):
        s="Elt:"
        for k in self.__hash.keys():
            s+=" %s=%s" % (k, self.__hash[k])
        return s

    def __cmp__(self,other):
        return cmp(self.__hash,other.__hash)

class EltSet:
    """A set of elements with <and> operator
    """
    def __init__(self, elts):
        self.elts = elts

    def __len__(self):
        return len(self.elts)

    def add_elt(self,e):
        self.elts.append(e)

    def __iter__(self):
        for i in range(len(self.elts)-1,-1,-1):
            yield(self.elts[i])

    def __repr__(self):
        s="{"
        for elt in self.elts:
            s+= "%r| " % elt.__repr__()
        s+="}"
        return s

    def merge_elts(self):
        if(self.elts!=[]):
            e = self.elts[0]
            for o in self.elts[1:len(self.elts)]:
                e.merge_elt(o)
            self.elts=[e]

class EltGrp:
    """Represents a network resource, list of IP or IP/mask
    (source or destination)
    """
    def __init__(self,ID,name,op):
        assert isinstance(ID, int)
        self.ID = ID
        self.op = op
        self.name = name
        self.link_t=[]
        self.eltset_t=[]
        self.eltgrp_t=[]
        self.elts=[]
        self.scpt = 0

    def add_eltset(self,eltset):
        self.eltset_t.append(eltset)

    def remove_eltset(self,eltset):
        self.eltset_t.remove(eltset)

    def add_elt(self,e):
        self.elts.append(e)

    def add_eltlink(self,lk):
        self.link_t.append(lk)

    def iter_eltsets(self):
        for ind in range(len(self.eltset_t)-1,-1,-1):
            yield self.eltset_t[ind]

    def has_link(self):
        if(self.link_t==[]):
            return False
        return True

    def is_ok(self):
        for es in (self.link_t):
            for elt in es.elts:
                if(elt.get('type')=="link"):
                    return False
        return True

    def is_solved(self):
        """Tells if the group of elements is solved
        """
        if(self.eltgrp_t == []):
            return True
        return False

    def resolv_or(self):
        """Put elements that come from links with or operator in elements list
        """
        if(self.eltset_t==[]):
            if(self.eltgrp_t!=[]):
                self.eltset_t = deepcopy(self.eltgrp_t[0].eltset_t) #init
                tmp = self.eltgrp_t[0]
                self.eltgrp_t.remove(tmp)

        for grp in self.eltgrp_t:
            if(len(self.eltset_t)==0):
                pass
            for eltset in grp.iter_eltsets():
                self.add_eltset(eltset)
        self.__del_links("or")
        self.op="or"

    def resolv(self):
        """Resolvs grpelts with <or> operators
           Then clean links groups
        """

        self.resolv_or()
        self.eltgrp_t=[]
        self.scpt+=1

    def base_norm(self,grplist,linklist):
        """Follow links elts with deep 1 and put the destination group elements
        """
        for es in (self.link_t):
            for elt in es.elts:
                if elt.get('type') != "link":
                    continue
                dest = elt.get('linkid')
                id = elt.get('ID')
                if dest:
                    grpdest = grplist.find(int(dest))
                else:
                    grpdest = None
                if grpdest is None:
                    raise LinkError(
                        "Link Error: elt=<%r> linkid=%r id=%r" \
                        % (elt, dest, id))
                gid = grpdest.ID
                assert isinstance(self.ID, int)
                assert isinstance(gid, int)
                if self.ID not in linklist:
                    linklist[self.ID]=set([gid])
                else:
                    #if gid not in linklist[self.ID]:
                    linklist[self.ID].add(gid)
                if gid in linklist:
                    if self.ID in linklist[gid]:
                        raise LoopError(
                            "Link Loop Detected in group ID=%s and elt %s: "
                            "dest %s already seen" % (self.ID, id, gid))
                self.eltgrp_t.append(grpdest)

        if self.ID in linklist:
            # Remove visited links
            for k in linklist.keys():
                lrem = linklist[self.ID]
                if self.ID in linklist[k]:
                    linklist[k].remove(self.ID)
                    linklist[k] |= lrem

    def norm(self,grplist,linklist,listsolved=None):
        """The recursive thing
        """
        if listsolved is None:
            listsolved = []
        if(self.has_link()):
            self.base_norm(grplist,linklist)
            for sub in self.eltgrp_t:
                sub.norm(grplist,linklist,listsolved=listsolved)
        if(not self.is_solved()):
            if listsolved.count(self.ID)==0:
                self.resolv()
            listsolved.append(self.ID)

    def __del_links(self,op):
        """delete GrpEltSets in links list wich have <op> operator
        """
        for grp in self.eltgrp_t:
            if grp.op == op:
                self.eltgrp_t.remove(grp)

    def __repr__(self):
        s = "EltGrp: <%s> <%s> <%s>\n" % (self.ID, self.name, self.op)
        for es in self.eltset_t:
            s += "\tEltSet <%r>\n" % es
        for link in self.link_t:
            s += "\tLink <%r>\n" % link
        return s

class EltGrpList:
    def __init__(self):
        self.__list=[]

    def add(self,eltgrp):
        self.__list.append(eltgrp)

    def find(self,index):
        assert isinstance(index, int)
        for a in self.__list:
            if (a.ID == index):
                return a
        return None

    def normall(self):
        l=[]
        l2={}
        for grp in self.__list:
            grp.norm(self,l2,listsolved=l)

    def get_eltgrps(self):
        return self.__list

    def __str__(self):
        return str(self.__list)

